# Copyright (c) 2012 Cloudera, Inc. All rights reserved.

from cmf.monitor.json_metrics_utils import ParseError
from cmf.throttling_logger import ThrottlingLogger
from cmf.monitor import schema
import logging

# Exhaustive list of all currently exposed Flume metrics, and their
# corresponding entry in our own schema.
_FLUME_JSON_METRICS = {
  "source": {
    "StartTime" : schema.FLUME_SOURCE_START_TIME,
    "StopTime" : schema.FLUME_SOURCE_STOP_TIME,
    "EventReceivedCount" : schema.FLUME_SOURCE_EVENT_RECEIVED_COUNT,
    "EventAcceptedCount" : schema.FLUME_SOURCE_EVENT_ACCEPTED_COUNT,
    "AppendReceivedCount" : schema.FLUME_SOURCE_APPEND_RECEIVED_COUNT,
    "AppendAcceptedCount" : schema.FLUME_SOURCE_APPEND_ACCEPTED_COUNT,
    "AppendBatchReceivedCount" : schema.FLUME_SOURCE_APPEND_BATCH_RECEIVED_COUNT,
    "AppendBatchAcceptedCount" : schema.FLUME_SOURCE_APPEND_BATCH_ACCEPTED_COUNT
  },
  "channel": {
    "StartTime" : schema.FLUME_CHANNEL_START_TIME,
    "StopTime" : schema.FLUME_CHANNEL_STOP_TIME,
    "ChannelSize" : schema.FLUME_CHANNEL_SIZE,
    "ChannelCapacity" : schema.FLUME_CHANNEL_CAPACITY,
    "EventPutAttemptCount" : schema.FLUME_CHANNEL_EVENT_PUT_ATTEMPT_COUNT,
    "EventTakeAttemptCount" : schema.FLUME_CHANNEL_EVENT_TAKE_ATTEMPT_COUNT,
    "EventPutSuccessCount" : schema.FLUME_CHANNEL_EVENT_PUT_SUCCESS_COUNT,
    "EventTakeSuccessCount" : schema.FLUME_CHANNEL_EVENT_TAKE_SUCCESS_COUNT
  },
  "sink": {
    "StartTime" : schema.FLUME_SINK_START_TIME,
    "StopTime" : schema.FLUME_SINK_STOP_TIME,
    "ConnectionCreatedCount" : schema.FLUME_SINK_CONNECTION_CREATED_COUNT,
    "ConnectionClosedCount" : schema.FLUME_SINK_CONNECTION_CLOSED_COUNT,
    "ConnectionFailedCount" : schema.FLUME_SINK_CONNECTION_FAILED_COUNT,
    "BatchEmptyCount" : schema.FLUME_SINK_BATCH_EMPTY_COUNT,
    "BatchUnderflowCount" : schema.FLUME_SINK_BATCH_UNDERFLOW_COUNT,
    "BatchCompleteCount" : schema.FLUME_SINK_BATCH_COMPLETE_COUNT,
    "EventDrainAttemptCount" : schema.FLUME_SINK_EVENT_DRAIN_ATTEMPT_COUNT,
    "EventDrainSuccessCount" : schema.FLUME_SINK_EVENT_DRAIN_SUCCESS_COUNT
  }
}

LOG = logging.getLogger(__name__)
THROTTLED_LOG = ThrottlingLogger(LOG)

class JsonParser(object):
  def __init__(self, json_dict):
    """
    May raise ParseError if the json structure is not what we expect.
    """
    self._json = json_dict
    self.updates = {}

    for (prefixed_name, component_metrics) in json_dict.items():
      if not "Type" in component_metrics:
        raise ParseError("Missing type for component '%s'" % prefixed_name)
      component_type = component_metrics["Type"].lower()
      prefix = component_type + "."
      if not prefixed_name.lower().startswith(prefix):
        raise ParseError("Expected '%s' prefix in component name '%s'" % \
            (prefix, prefixed_name))
      if (len(prefixed_name) <= len(prefix)):
        raise ParseError("No component name after type prefix '%s'" % \
            prefixed_name)
      component_name   = prefixed_name[len(prefix):]
      component_schema = _FLUME_JSON_METRICS.get(component_type)
      if not component_schema:
        THROTTLED_LOG.warning("Unknown type '%s' for component '%s'" % \
            (component_type, component_name))
        continue
      metricdict = {}
      for (metric_json_key, metric_id) in component_schema.items():
        if metric_json_key in component_metrics:
          metricdict[metric_id] = long(component_metrics.get(metric_json_key))
        else:
          raise ParseError("Flume metric '%s' not found for component '%s'" % \
              (metric_json_key, component_name))
      self.updates[(component_type, component_name)] = metricdict

#
# This is what the Flume metrics JSON looks like (pretty-printed for
# readability):
#
SAMPLE = r"""
{
   "SOURCE.source2":{
      "AppendBatchAcceptedCount":"0",
      "AppendBatchReceivedCount":"0",
      "Type":"SOURCE",
      "EventAcceptedCount":"0",
      "AppendReceivedCount":"0",
      "StopTime":"0",
      "EventReceivedCount":"0",
      "StartTime":"1345147577837",
      "AppendAcceptedCount":"0"
   },
   "SINK.sink1":{
      "BatchCompleteCount":"0",
      "ConnectionFailedCount":"8",
      "EventDrainAttemptCount":"0",
      "ConnectionCreatedCount":"0",
      "BatchEmptyCount":"0",
      "Type":"SINK",
      "ConnectionClosedCount":"0",
      "EventDrainSuccessCount":"0",
      "StopTime":"0",
      "StartTime":"1345147577389",
      "BatchUnderflowCount":"0"
   },
   "SINK.test":{
      "BatchCompleteCount":"0",
      "ConnectionFailedCount":"0",
      "EventDrainAttemptCount":"0",
      "ConnectionCreatedCount":"0",
      "BatchEmptyCount":"5",
      "Type":"SINK",
      "ConnectionClosedCount":"0",
      "EventDrainSuccessCount":"0",
      "StopTime":"0",
      "StartTime":"1345147577400",
      "BatchUnderflowCount":"0"
   },
   "CHANNEL.channel1":{
      "EventPutSuccessCount":"10000",
      "Type":"CHANNEL",
      "StopTime":"0",
      "EventPutAttemptCount":"10011",
      "ChannelSize":"10000",
      "StartTime":"1345147577389",
      "EventTakeSuccessCount":"0",
      "ChannelCapacity":"10000",
      "EventTakeAttemptCount":"0"
   },
   "CHANNEL.test":{
      "EventPutSuccessCount":"0",
      "Type":"CHANNEL",
      "StopTime":"0",
      "EventPutAttemptCount":"0",
      "ChannelSize":"0",
      "StartTime":"1345147577389",
      "EventTakeSuccessCount":"0",
      "ChannelCapacity":"10000",
      "EventTakeAttemptCount":"6"
   }
}
"""

# The correct parsing of the above example JSON.
SAMPLE_CORRECT_UPDATES = {
  ('source' , 'source2' ): {11722: 0L, 11723: 0L, 11724: 0L, 11725: 0L, 11726: 0L, 11727: 0L, 11728: 1345147577837L, 11729: 0L},
  ('channel', 'channel1'): {11704: 10000L, 11705: 10000L, 11706: 10011L, 11707: 0L, 11708: 10000L, 11709: 0L, 11710: 1345147577389L, 11711: 0L},
  ('channel', 'test'    ): {11704:     0L, 11705: 10000L, 11706:     0L, 11707: 6L, 11708:     0L, 11709: 0L, 11710: 1345147577389L, 11711: 0L},
  ('sink'   , 'sink1'   ): {11712: 0L, 11713: 0L, 11714: 8L, 11715: 0L, 11716: 0L, 11717: 0L, 11718: 0L, 11719: 0L, 11720: 1345147577389L, 11721: 0L},
  ('sink'   , 'test'    ): {11712: 0L, 11713: 0L, 11714: 0L, 11715: 5L, 11716: 0L, 11717: 0L, 11718: 0L, 11719: 0L, 11720: 1345147577400L, 11721: 0L}
}

# Below is the flume configuration that produced the output above.
"""
tier1.sources = source1 source2
tier1.channels = channel1 test
tier1.sinks = sink1 test

tier1.channels.channel1.type = memory
tier1.channels.test.type = memory
tier1.channels.channel1.capacity = 10000
tier1.channels.test.capacity = 10000
tier1.channels.channel1.transactionCapacity = 1000
tier1.channels.test.transactionCapacity = 1000

tier1.sinks.sink1.channel = channel1
tier1.sinks.sink2.channel = test
tier1.sinks.test.channel = test
tier1.sources.source1.channels = channel1
tier1.sources.source2.channels = test

tier1.sources.source1.type = seq
tier1.sources.source2.type = avro
tier1.sinks.sink1.type = avro
tier1.sinks.sink2.type = logger
tier1.sinks.test.type = hdfs

tier1.sources.source2.port = 9099
tier1.sources.source2.bind = 0.0.0.0

tier1.sinks.sink1.hostname = centos57-14.ent.cloudera.com
tier1.sinks.sink1.port = 9099

tier1.sinks.test.hdfs.path = hdfs://blah.ent.cloudera.com:8020/flume
tier1.sinks.test.hdfs.filePrefix = testHDFSsink
tier1.sinks.test.hdfs.rollCount=0
tier1.sinks.test.hdfs.rollSize=33333333
tier1.sinks.test.hdfs.batchSize = 1000
tier1.sinks.test.hdfs.txnEventMax = 1000
"""
