# Copyright (c) 2012 Cloudera, Inc. All rights reserved.

import logging
import time
import re

import psutil

from cmf.monitor.host.disk_device import DiskDevice
from cmf.throttling_logger import ThrottlingLogger

LOG = logging.getLogger(__name__)
THROTTLED_LOG = ThrottlingLogger(LOG)

class DiskDevices(object):
  _DEVICE_RE = re.compile("^/dev/((\D+)\d*)$")
  _DEVICE_EXCLUDE_PATTERN_DEFAULT = "^$"
  _DEVICE_EXCLUDE_REGEX_DEFAULT = re.compile(_DEVICE_EXCLUDE_PATTERN_DEFAULT)

  """Provides access to all the mounted disk devices on the host"""
  def __init__(self):
    self._disks = []
    self._latest_updates_pair = ([], None)
    self._last_reported_timestamp = 0
    self._device_exclude_regex = self._DEVICE_EXCLUDE_REGEX_DEFAULT
    self._last_regex_pattern = self._DEVICE_EXCLUDE_PATTERN_DEFAULT

    partitions = []
    try:
      partitions = psutil.disk_partitions()
    except Exception, ex:
      LOG.exception("Error getting disk partitions: %s" % (ex,))

    # list of DiskDevice objects for all disk devices and partitions
    disks = []
    # keep a set of the disks (e.g. sda, not partitions) already added to avoid duplicates
    disks_added_set = set()
    for partition in partitions:
      try:
        # Extracts the device and partition from the full device name:
        # /dev/sda1 matches groups ('sda1', 'sda')
        groups = self._DEVICE_RE.match(partition.device).groups()
        assert len(groups) == 2, "Unexpected device name '%s'" % (self._device)
        # disk_name for /dev/sda1 will be sda
        disk_name = groups[1]
        # partition_name for /dev/sda1 will be sda1
        partition_name = groups[0]

        if disk_name not in disks_added_set:
          try:
            disks.append(DiskDevice(disk_name))
            disks_added_set.add(disk_name)
            LOG.debug("Added DiskDevice for disk=%s" % disk_name)
          except Exception, ex:
            LOG.warning("Failed to add DiskDevice for disk='%s'" % (disk_name))

        # It's possible that psutil returns a "partition" object that is just a device
        if disk_name != partition_name:
          try:
            disks.append(DiskDevice(disk_name, partition_name))
            LOG.debug("Added DiskDevice for partition=%s" % partition_name)
          except Exception, ex:
            LOG.warning("Failed to add DiskDevice for partition='%s'" % (partition_name))

      except Exception, ex:
        LOG.exception("Error creating disk devices '%s'" % (partition.device))

    self._disks = disks

  def set_device_exclude_regex(self, regex_pattern):
    if self._last_regex_pattern != regex_pattern:
      self._last_regex_pattern = regex_pattern
      try:
        self._device_exclude_regex = re.compile(regex_pattern)
      except re.error:
        THROTTLED_LOG.error("Unable to compile disk device exclusion regex '%s', using default.",
                            regex_pattern)
        self._device_exclude_regex = self._DEVICE_EXCLUDE_REGEX_DEFAULT

  def collect(self):
    updates = []
    # capture the regex as a local variable in case it changes
    exclude_regex = self._device_exclude_regex

    for disk in self._disks:
      try:
        if not exclude_regex.match(disk.device()):
          update = disk.collect_update()
          if update is not None and update.contains_metrics():
            updates.append(update)
      except Exception:
        THROTTLED_LOG.exception("Error collecting disk update for %s" % (disk.device(),))

    self._latest_updates_pair = (updates, time.time())

  def get_disk_updates(self):
    (updates, timestamp) = self._latest_updates_pair

    if timestamp == self._last_reported_timestamp:
      return list()

    self._last_reported_timestamp = timestamp
    return updates