# Copyright (c) 2011-2012 Cloudera, Inc. All rights reserved.

import logging
import time

from cmf.monitor import schema
from cmf.monitor.abstract_hdfs_monitor import AbstractHdfsMonitor
from cmf.monitor.base import DebugStatusReport
from cmf.monitor.journalnode.json_parser_jmx import JsonParser
from cmf.monitor.json_metrics_utils import ParseError, add_jvm_metrics
from cmf.monitor.proc_metrics_utils import add_proc_metrics
from cmf.monitor.role_update import RoleUpdate

LOG = logging.getLogger('JournalNodeMonitor')

_ROLETYPE = 'JOURNALNODE'

# Section and key names in cloudera-monitor.properties
# Since CM 4.1
_SECTION = 'journalnode'
_SERVICE_NAME = 'service_name'
_ROLE_NAME = 'role_name'
_HOST = 'host'
_DFS_JOURNALNODE_HTTP_PORT = 'dfs_journalnode_http_port'
_DFS_JOURNALNODE_EDITS_DIR = 'dfs_journalnode_edits_dir'
_LOG_DIR = 'log_dir'
_COLLECT_INTERVAL = 'collect_interval'
_SERVICE_VERSION = 'service_version'

class JournalNodeMonitor(AbstractHdfsMonitor):

  def __init__(self, daemon, registry):
    AbstractHdfsMonitor.__init__(self, 
                                 daemon,
                                 registry,
                                 _SECTION,
                                 _DFS_JOURNALNODE_HTTP_PORT)
    self._last_conf_path = None
    self._metrics_tuple = (None, None, None)
    self._last_reported_timestamp = None
    self._last_collection_ended = None
    self._last_collection_length = None
    self._role_dir_monitors = [ ]

    # Config data
    self._name = None
    self._service_name = None

  def __str__(self):
    return "JournalNodeMonitor for %s" % (self._name,)

  def _reload_config(self, confparse):
    """Returns True/False for success"""
    try:
      # Extraction
      name = confparse.get(_SECTION, _ROLE_NAME)
      service_name = confparse.get(_SECTION, _SERVICE_NAME)
      service_version = confparse.getint(_SECTION, _SERVICE_VERSION)
      dfs_journalnode_edits_dir = \
          confparse.get(_SECTION, _DFS_JOURNALNODE_EDITS_DIR)
      host = confparse.get(_SECTION, _HOST)
      web_port = confparse.getint(_SECTION, _DFS_JOURNALNODE_HTTP_PORT)
      log_dir = confparse.get(_SECTION, _LOG_DIR)
    except ValueError, ex:
      self.record_error(ex)
      LOG.error("Value error in %s: %s" % (confparse.path, ex))
      return False

    metrics_url = self._compute_metrics_url(service_version, host, web_port)
    if metrics_url is None:
      return False

    # Assign
    self._name = name
    self._service_name = service_name
    self._metrics_url = metrics_url

    # Make monitors for the log directory and the data directories.
    self._role_dir_monitors = [ ]
    self.create_dir_monitor(log_dir, self._role_dir_monitors)
    self.create_dir_monitor(dfs_journalnode_edits_dir, self._role_dir_monitors)

    return True

  def collect(self):
    """Overrides AbstractMonitor. Collect metrics. Invoked asynchronously."""
    start = time.time()

    # Reload config if necessary
    confparse = self.conf_parser
    if self._last_conf_path != confparse.path:
      if not self._reload_config(confparse):
        return
      self._last_conf_path = confparse.path

    # Collect directory info
    for mon in self._role_dir_monitors:
      mon.collect()

    # If our pid value is None, it means that the associated process is not
    # running. If this is the case, we skip metric collection.
    if self._pid is None:
      self._metrics_tuple = (None, None, None)
      return

    LOG.debug("Fetching JournalNode metrics at '%s'" % (self._metrics_url,))
    self._metrics_tuple = self.collect_metrics_from_url(self._metrics_url)

    self._last_collection_ended = time.time()
    self._last_collection_length = self._last_collection_ended - start

  def parse_metrics(self, json):
    return JsonParser(json)

  def get_role_update(self):
    """Returns a RoleUpdate to be sent to the firehose"""
    update = RoleUpdate(name=self._name,
                        roletype=_ROLETYPE,
                        service_name=self._service_name)

    if self.add_metrics_from_webserver(self._metrics_tuple,
                                       self._last_reported_timestamp,
                                       update):
      self._last_reported_timestamp = update.timestamp
                                    
    add_proc_metrics(update, self._pid, LOG)

    for mon in self._role_dir_monitors:
      update.add_dir_update(mon.get_role_dir_update())

    return update

  def add_webserver_metrics_to_role_update(self, update, metrics):
    add_jvm_metrics(update, metrics, True)

    if metrics.jvm is not None:
      update.add_metric(schema.LOG_ERROR,
                        metrics.jvm.LogError)
      update.add_metric(schema.LOG_FATAL,
                        metrics.jvm.LogFatal)
      update.add_metric(schema.LOG_INFO,
                        metrics.jvm.LogInfo)
      update.add_metric(schema.LOG_WARN,
                        metrics.jvm.LogWarn)
    if metrics.detailed_rpc is not None:
      update.add_metric(schema.ACCEPT_RECOVERY_AVG_TIME,
                        metrics.detailed_rpc.AcceptRecoveryAvgTime)
      update.add_metric(schema.ACCEPT_RECOVERY_NUM_OPS,
                        metrics.detailed_rpc.AcceptRecoveryNumOps)
      update.add_metric(schema.FINALIZE_LOG_SEGMENT_AVG_TIME,
                        metrics.detailed_rpc.FinalizeLogSegmentAvgTime)
      update.add_metric(schema.FINALIZE_LOG_SEGMENT_NUM_OPS,
                        metrics.detailed_rpc.FinalizeLogSegmentNumOps)
      update.add_metric(schema.GET_EDIT_LOG_MANIFEST_AVG_TIME,
                        metrics.detailed_rpc.GetEditLogManifestAvgTime)
      update.add_metric(schema.GET_EDIT_LOG_MANIFEST_NUM_OPS,
                        metrics.detailed_rpc.GetEditLogManifestNumOps)
      update.add_metric(schema.GET_JOURNAL_STATE_AVG_TIME,
                        metrics.detailed_rpc.GetJournalStateAvgTime)
      update.add_metric(schema.GET_JOURNAL_STATE_NUM_OPS,
                        metrics.detailed_rpc.GetJournalStateNumOps)
      update.add_metric(schema.JOURNAL_AVG_TIME,
                        metrics.detailed_rpc.JournalAvgTime)
      update.add_metric(schema.JOURNAL_NUM_OPS,
                        metrics.detailed_rpc.JournalNumOps)
      update.add_metric(schema.NEW_EPOCH_AVG_TIME,
                        metrics.detailed_rpc.NewEpochAvgTime)
      update.add_metric(schema.NEW_EPOCH_NUM_OPS,
                        metrics.detailed_rpc.NewEpochNumOps)
      update.add_metric(schema.PREPARE_RECOVERY_AVG_TIME,
                        metrics.detailed_rpc.PrepareRecoveryAvgTime)
      update.add_metric(schema.PREPARE_RECOVERY_NUM_OPS,
                        metrics.detailed_rpc.PrepareRecoveryNumOps)
      update.add_metric(schema.START_LOG_SEGMENT_AVG_TIME,
                        metrics.detailed_rpc.StartLogSegmentAvgTime)
      update.add_metric(schema.START_LOG_SEGMENT_NUM_OPS,
                        metrics.detailed_rpc.StartLogSegmentNumOps)
    if metrics.ugi is not None:
      update.add_metric(schema.LOGIN_FAILURE_AVG_TIME,
                        metrics.ugi.LoginFailureAvgTime)
      update.add_metric(schema.LOGIN_FAILURE_NUM_OPS,
                        metrics.ugi.LoginFailureNumOps)
      update.add_metric(schema.LOGIN_SUCCESS_AVG_TIME,
                        metrics.ugi.LoginSuccessAvgTime)
      update.add_metric(schema.LOGIN_SUCCESS_NUM_OPS,
                        metrics.ugi.LoginSuccessNumOps)
    if metrics.journal is not None:
      update.add_metric(schema.BATCHES_WRITTEN,
                        metrics.journal.BatchesWritten)
      update.add_metric(schema.BATCHES_WRITTEN_WHILE_LAGGING,
                        metrics.journal.BatchesWrittenWhileLagging)
      update.add_metric(schema.BYTES_WRITTEN,
                        metrics.journal.BytesWritten)
      update.add_metric(schema.CURRENT_LAG_TXNS_JOURNALNODE,
                        metrics.journal.CurrentLagTxns)
      update.add_metric(schema.LAST_PROMISED_EPOCH,
                        metrics.journal.LastPromisedEpoch)
      update.add_metric(schema.LAST_WRITER_EPOCH,
                        metrics.journal.LastWriterEpoch)
      update.add_metric(schema.LAST_WRITTEN_TX_ID,
                        metrics.journal.LastWrittenTxId)
      update.add_metric(schema.SYNCS300S50TH_PERCENTILE_LATENCY_MICROS,
                        metrics.journal.Syncs300s50thPercentileLatencyMicros)
      update.add_metric(schema.SYNCS300S75TH_PERCENTILE_LATENCY_MICROS,
                        metrics.journal.Syncs300s75thPercentileLatencyMicros)
      update.add_metric(schema.SYNCS300S90TH_PERCENTILE_LATENCY_MICROS,
                        metrics.journal.Syncs300s90thPercentileLatencyMicros)
      update.add_metric(schema.SYNCS300S95TH_PERCENTILE_LATENCY_MICROS,
                        metrics.journal.Syncs300s95thPercentileLatencyMicros)
      update.add_metric(schema.SYNCS300S99TH_PERCENTILE_LATENCY_MICROS,
                        metrics.journal.Syncs300s99thPercentileLatencyMicros)
      update.add_metric(schema.SYNCS300S_NUM_OPS,
                        metrics.journal.Syncs300sNumOps)
      update.add_metric(schema.SYNCS3600S50TH_PERCENTILE_LATENCY_MICROS,
                        metrics.journal.Syncs3600s50thPercentileLatencyMicros)
      update.add_metric(schema.SYNCS3600S75TH_PERCENTILE_LATENCY_MICROS,
                        metrics.journal.Syncs3600s75thPercentileLatencyMicros)
      update.add_metric(schema.SYNCS3600S90TH_PERCENTILE_LATENCY_MICROS,
                        metrics.journal.Syncs3600s90thPercentileLatencyMicros)
      update.add_metric(schema.SYNCS3600S95TH_PERCENTILE_LATENCY_MICROS,
                        metrics.journal.Syncs3600s95thPercentileLatencyMicros)
      update.add_metric(schema.SYNCS3600S99TH_PERCENTILE_LATENCY_MICROS,
                        metrics.journal.Syncs3600s99thPercentileLatencyMicros)
      update.add_metric(schema.SYNCS3600S_NUM_OPS,
                        metrics.journal.Syncs3600sNumOps)
      update.add_metric(schema.SYNCS60S50TH_PERCENTILE_LATENCY_MICROS,
                        metrics.journal.Syncs60s50thPercentileLatencyMicros)
      update.add_metric(schema.SYNCS60S75TH_PERCENTILE_LATENCY_MICROS,
                        metrics.journal.Syncs60s75thPercentileLatencyMicros)
      update.add_metric(schema.SYNCS60S90TH_PERCENTILE_LATENCY_MICROS,
                        metrics.journal.Syncs60s90thPercentileLatencyMicros)
      update.add_metric(schema.SYNCS60S95TH_PERCENTILE_LATENCY_MICROS,
                        metrics.journal.Syncs60s95thPercentileLatencyMicros)
      update.add_metric(schema.SYNCS60S99TH_PERCENTILE_LATENCY_MICROS,
                        metrics.journal.Syncs60s99thPercentileLatencyMicros)
      update.add_metric(schema.SYNCS60S_NUM_OPS,
                        metrics.journal.Syncs60sNumOps)
      update.add_metric(schema.TXNS_WRITTEN,
                        metrics.journal.TxnsWritten)
    if metrics.rpc is not None:
      update.add_metric(schema.RPC_CALL_QUEUE_LENGTH,
                        metrics.rpc.CallQueueLength)
      update.add_metric(schema.RPC_NUM_OPEN_CONNECTIONS,
                        metrics.rpc.NumOpenConnections)
      update.add_metric(schema.RPC_RECEIVED_BYTES,
                        metrics.rpc.ReceivedBytes)
      update.add_metric(schema.RPC_AUTHENTICATION_FAILURES,
                        metrics.rpc.RpcAuthenticationFailures)
      update.add_metric(schema.RPC_AUTHENTICATION_SUCCESSES,
                        metrics.rpc.RpcAuthenticationSuccesses)
      update.add_metric(schema.RPC_AUTHORIZATION_FAILURES,
                        metrics.rpc.RpcAuthorizationFailures)
      update.add_metric(schema.RPC_AUTHORIZATION_SUCCESSES,
                        metrics.rpc.RpcAuthorizationSuccesses)
      update.add_metric(schema.RPC_PROCESSING_TIME_AVG_TIME,
                        metrics.rpc.RpcProcessingTimeAvgTime)
      update.add_metric(schema.RPC_PROCESSING_TIME_NUM_OPS,
                        metrics.rpc.RpcProcessingTimeNumOps)
      update.add_metric(schema.RPC_QUEUE_TIME_AVG_TIME,
                        metrics.rpc.RpcQueueTimeAvgTime)
      update.add_metric(schema.RPC_QUEUE_TIME_NUM_OPS,
                        metrics.rpc.RpcQueueTimeNumOps)
      update.add_metric(schema.RPC_SENT_BYTES,
                        metrics.rpc.SentBytes)
    if metrics.metrics_stats is not None:
      update.add_metric(schema.METRICS_DROPPED_PUB_ALL,
                        metrics.metrics_stats.DroppedPubAll)
      update.add_metric(schema.METRICS_NUM_ACTIVE_SINKS,
                        metrics.metrics_stats.NumActiveSinks)
      update.add_metric(schema.METRICS_NUM_ACTIVE_SOURCES,
                        metrics.metrics_stats.NumActiveSources)
      update.add_metric(schema.METRICS_NUM_ALL_SINKS,
                        metrics.metrics_stats.NumAllSinks)
      update.add_metric(schema.METRICS_NUM_ALL_SOURCES,
                        metrics.metrics_stats.NumAllSources)
      update.add_metric(schema.METRICS_PUBLISH_AVG_TIME,
                        metrics.metrics_stats.PublishAvgTime)
      update.add_metric(schema.METRICS_PUBLISH_NUM_OPS,
                        metrics.metrics_stats.PublishNumOps)
      update.add_metric(schema.METRICS_SNAPSHOT_AVG_TIME,
                        metrics.metrics_stats.SnapshotAvgTime)
      update.add_metric(schema.METRICS_SNAPSHOT_NUM_OPS,
                        metrics.metrics_stats.SnapshotNumOps)

  def get_debug_status_report(self):
    return DebugStatusReport(config_items=self.conf_parser.items(_SECTION),
                             errors=(self.last_error,),
                             last_collection_ended=self._last_collection_ended,
                             last_collection_length=self._last_collection_length)

  def get_subject_type(self):
    return "JOURNALNODE"
